"""
Django cache backend.
"""


class DjangoCache:
    """
    Allows Site Configuration client to configure caching with Django
    """
    def __init__(self, cache_name, cache_timeout=300):
        self.cache_name = cache_name
        self.cache_timeout = cache_timeout
        self._django_cache = None

    def get_django_cache(self):
        """
        Lazily instantiate Django cache to avoid `ImproperlyConfigured` error.
        """
        if not self._django_cache:
            from django.core.cache import caches
            self._django_cache = caches[self.cache_name]
        return self._django_cache

    def set(self, key, value):
        kwargs = {}
        if self.cache_timeout is not None:
            kwargs['timeout'] = self.cache_timeout
        return self.get_django_cache().set(key, value, **kwargs)

    def get(self, key):
        return self.get_django_cache().get(key)

    def delete(self, key):
        return self.get_django_cache().delete(key)



"""
Google cloud read-only storage.
"""
from logging import getLogger

from google.cloud import storage
from google.api_core.exceptions import NotFound

log = getLogger(__name__)


class GoogleCloudStorage:
    """
    Allows Site Configuration Client to read configurations from
    Google Cloud Storage bucket
    """
    def __init__(self, bucket_name):
        self.bucket_name = bucket_name
        self.client = storage.Client()
        self.bucket = self.client.get_bucket(self.bucket_name)

    def read(self, file_path):
        """
        Gets the text content of a cloud storage bucket.
        If the file don't exists, it returns `None`.
        """
        blob = self.bucket.get_blob(file_path)
        if blob:
            try:
                return blob.download_as_bytes().decode('utf-8')
            except NotFound:
                pass

        log.info('File path not found: %s', file_path)
        return None



"""
An API adapter and in-memory cache for the Site Configuration backend configuration.
"""

from django.conf import settings

from dateutil import parser


AMC_V1_STRUCTURE_VERSION = 'amc-v1'


class SiteConfigAdapter:
    """
    Adapter for Open edX translates the values in a format that Open edX can use.
    """

    backend_configs = None

    TYPE_SETTING = 'setting'
    TYPE_SECRET = 'secret'  # nosec
    TYPE_ADMIN = 'admin'
    TYPE_PAGE = 'page'
    TYPE_CSS = 'css'

    def __init__(self, site_uuid, status='live'):
        self.site_uuid = site_uuid
        self.status = status

    def get_backend_configs(self):
        if not self.backend_configs:
            client = settings.SITE_CONFIG_CLIENT
            self.backend_configs = client.get_backend_configs(self.site_uuid, self.status)
        return self.backend_configs

    def delete_backend_configs_cache(self):
        """
        Enforce getting a fresh entry for the current context/request and following ones.
        """
        self.backend_configs = None
        client = settings.SITE_CONFIG_CLIENT
        client.delete_cache_for_site(self.site_uuid, self.status)

    def get_value_of_type(self, config_type, name, default):
        all_configs = self.get_backend_configs()['configuration']
        type_configs = all_configs[config_type]
        return type_configs.get(name, default)

    def get_css_variables_dict(self):
        """
        Get a variable_name:value dictionary of CSS variables.
        """
        config = self.get_backend_configs()['configuration']
        return config[self.TYPE_CSS]

    def get_site_info(self):
        """
        Get the site information from site config.

        Return dict(
          always_active: boolean
          domain_name: string
          environment: dict(name: string)
          subscription_ends: date
          tier: string
          uuid: UUID
        )

        Note, is_active is not returned from the response since it can be outdated.
        """
        site_info = self.get_backend_configs()['site'].copy()  # Get a copy of the object to manipulate it.

        # Remove `is_active` property to avoid caching it
        site_info.pop('is_active', None)

        # Convert the string to useful datetime object
        site_info['subscription_ends'] = parser.parse(site_info['subscription_ends'])

        return site_info



"""
Tests for site_config_client.openedx.test_helpers
"""
import pytest

try:
    from site_config_client.openedx.test_helpers import override_site_config
    from site_config_client.openedx import api as openedx_api
except ImportError:
    # Silent import failures for non-Open edX environments.
    pass


@pytest.mark.openedx
def test_secret_override():
    """
    Use `override_site_config` for a `secret` site configuration.
    """
    with override_site_config('secret', EMAIL_PASSWORD='test'):
        assert openedx_api.get_secret_value('EMAIL_PASSWORD') == 'test'


@pytest.mark.openedx
def test_without_overrides_fails():
    """
    Test that api fails without overrides due to missing packages in test environment.
    """

    with pytest.raises(NameError, match='name .configuration_helpers. is not defined'):
        openedx_api.get_secret_value('EMAIL_PASSWORD')



"""
Models for the Site Configuration client.
"""

from django.db import models


class SiteConfigClientEnabled(models.Model):
    """
    A model to enable the site configuration client for a specific site.

    This is mostly used by Open edX, so it's mostly useless for non Open edX projects.
    """

    class Meta:
        app_label = 'site_config_client'

    site_uuid = models.UUIDField(verbose_name='UUID of the site the client enabled for.', unique=True)
    note = models.CharField(verbose_name='Note for the the admin.', max_length=255)



"""
Test helpers for other packages to override testers.
"""

from unittest.mock import patch, Mock


def override_site_config(config_type, **config_overrides):
    """
    Override site config settings for a specific type of configuration.
    """
    def overrider_for_get_value_of_type(name, default=None, site_configuration=None):
        return config_overrides.get(name, default)

    return patch(
        'site_config_client.openedx.api.get_{type}_value'.format(type=config_type),
        Mock(side_effect=overrider_for_get_value_of_type),
    )



"""
Mocks for openedx.core.djangoapps.plugins.constants.

Mocks to make the plugin app works during tests
without being concerned to actually test the plugin configs.
"""

from unittest.mock import Mock

# Minimal mocks to reduce test maintenance costs.
SettingsType = Mock()
ProjectType = Mock()
PluginSettings = Mock()



"""
Test SiteConfigApp
"""
import pytest

try:
    import openedx  # noqa
    OPENEDX_ENVIRONMENT = True
except ImportError:
    OPENEDX_ENVIRONMENT = False


@pytest.mark.openedx
def test_plugin_config_openedx():
    """
    Check for syntax or other severe errors in SiteConfigApp.plugin_app
    """
    # Local import to avoid test failures for non-openedx tests
    from site_config_client import apps

    config = apps.SiteConfigApp('siteconfig', apps)
    assert type(config.plugin_app) == dict, 'Should have Open edX configs.'


@pytest.mark.django
@pytest.mark.skipif(OPENEDX_ENVIRONMENT, reason='Should not run if `openedx` is available')
def test_plugin_config_django():
    """
    Check for syntax or other severe errors in SiteConfigApp.plugin_app
    """
    # Local import to avoid test failures for non-django tests
    from site_config_client import apps

    config = apps.SiteConfigApp('siteconfig', apps)
    assert not getattr(config, 'plugin_app', None), 'Plugin app should only be available in Open edX environment.'



"""
External Open edX Python API helpers goes here.

### API Contract:
 * Those APIs should be stable and abstract internal changes.

 * Non-stable and internal APIs should be placed in other modules.

 * The parameters of existing functions should change in a backward compatible way:
   - No parameters should be removed from the function
   - New parameters should have safe defaults
 * For breaking changes, new functions should be created
"""

try:
    from openedx.core.djangoapps.site_configuration import helpers as configuration_helpers
except ImportError:
    # Silence the initial import error, but runtime errors will occur in tests and non-Open edX environments.
    # In tests, `configuration_helpers` can be mocked.
    pass


def get_current_configuration():
    """
    Gets current site configuration.
    """
    return configuration_helpers.get_current_site_configuration()


def get_admin_value(name, default=None, site_configuration=None):
    """
    Get `admin` setting from the site configuration service.

    Proxy for `site_configuration.get_admin_setting` until site_configuration is deprecated.
    """
    if not site_configuration:
        site_configuration = get_current_configuration()

    if site_configuration:
        return site_configuration.get_admin_setting(name, default)
    return default


def get_secret_value(name, default=None, site_configuration=None):
    """
    Get `secret` setting from the site configuration service.

    Proxy for `site_configuration.get_secret_value` until site_configuration is deprecated.
    """
    if not site_configuration:
        site_configuration = get_current_configuration()

    if site_configuration:
        return site_configuration.get_secret_value(name, default)
    return default


def get_setting_value(name, default=None, site_configuration=None):
    """
    Get `setting` setting from the site configuration service.

    Proxy for `site_configuration.get_value` until site_configuration is deprecated.
    """
    if not site_configuration:
        site_configuration = get_current_configuration()

    if site_configuration:
        return site_configuration.get_value(name, default)
    return default


def get_page_value(name, default=None, site_configuration=None):
    """
    Get `page` setting from the site configuration service.

    Proxy for `site_configuration.get_page_content` until site_configuration is deprecated.
    """
    if not site_configuration:
        site_configuration = get_current_configuration()

    if site_configuration:
        return site_configuration.get_page_content(name, default)
    return default



# Generated by Django 2.2.24 on 2022-01-13 04:22

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SiteConfigClientEnabled',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('site_uuid', models.UUIDField(unique=True, verbose_name='UUID of the site the client enabled for.')),
                ('note', models.CharField(max_length=255, verbose_name='Note for the the admin.')),
            ],
        ),
    ]



"""
Tests for the openedx.api module.
"""
from unittest.mock import patch, Mock

import pytest

try:
    from site_config_client.openedx import api as openedx_api
except ImportError:
    # Silent import failures for non-Open edX environments.
    pass


def with_current_configs(current_config):
    """
    @patch `get_current_site_configuration()`
    """
    configuration_helpers = Mock()
    configuration_helpers.get_current_site_configuration.return_value = current_config
    return patch(
        'site_config_client.openedx.api.configuration_helpers',
        configuration_helpers,
        create=True,
    )


@pytest.mark.openedx
def test_get_admin_value():
    """
    Test `get_admin_value()` helper for `admin` type of configurations.
    """
    current_config = Mock()
    current_config.get_admin_setting.return_value = 'password'
    with with_current_configs(current_config):
        admin_value = openedx_api.get_admin_value('IDP_CLIENT', 'default-client')
    assert admin_value == 'password'
    current_config.get_admin_setting.assert_called_with('IDP_CLIENT', 'default-client')


@pytest.mark.openedx
def test_get_secret_value():
    """
    Test `get_secret_value()` helper for `secret` type of configurations.
    """
    current_config = Mock()
    current_config.get_secret_value.return_value = 'password'
    with with_current_configs(current_config):
        secret_value = openedx_api.get_secret_value('EMAIL_PASSWORD', 'default-pass')
    assert secret_value == 'password'
    current_config.get_secret_value.assert_called_with('EMAIL_PASSWORD', 'default-pass')


@pytest.mark.openedx
def test_get_setting_value():
    """
    Test `get_setting_value()` helper for `setting` type of configurations.
    """
    current_config = Mock()
    current_config.get_value.return_value = 'pre-defined-site.com'
    with with_current_configs(current_config):
        setting = openedx_api.get_setting_value('SITE_NAME', 'defaultsite.com')
    assert setting == 'pre-defined-site.com'
    current_config.get_value.assert_called_with('SITE_NAME', 'defaultsite.com')


@pytest.mark.openedx
def test_get_page_value():
    """
    Test `get_page_value()` helper for `page` type of configurations.
    """
    current_config = Mock()
    current_config.get_page_content.return_value = '{"title": "About page"}'
    with with_current_configs(current_config):
        page_value = openedx_api.get_page_value('about', {})
    assert page_value == '{"title": "About page"}'
    current_config.get_page_content.assert_called_with('about', {})


@pytest.mark.openedx
@with_current_configs(current_config=None)  # Simulate an environment with no current configuration.
def test_use_default_on_missing_site_configuration():
    """
    Ensure all openedx_api helpers return `default` if `site_configuration` is missing.

    This is useful to return `None` for tests and when running on Open edX's main site.
    """
    assert openedx_api.get_secret_value('TEST_SECRET', 'default_secret') == 'default_secret'
    assert openedx_api.get_admin_value('TEST_CONFIG', 'dummy-default') == 'dummy-default'
    assert openedx_api.get_page_value('TEST_PAGE', 'dummy_page_content') == 'dummy_page_content'
    assert openedx_api.get_setting_value('CUSTOMER_ID', 'dummy-customer') == 'dummy-customer'



from unittest import mock
import pytest


@mock.patch('google.cloud.storage.Client', mock.Mock())
@pytest.mark.openedx
def test_plugin_production_settings():
    # Local import to avoid test failures for non-openedx tests
    from site_config_client.openedx.settings.production import plugin_settings

    settings = mock.Mock(
        SITE_CONFIG_CLIENT=None,
        SITE_CONFIG_CACHE_NAME='default',
        SITE_CONFIG_CACHE_TIMEOUT=3600,
        SITE_CONFIG_BASE_URL="http://service",
        SITE_CONFIG_ENVIRONMENT="staging",
        SITE_CONFIG_API_TOKEN="some-token",
        SITE_CONFIG_READ_ONLY_BUCKET="random-bucket",
    )

    plugin_settings(settings)

    assert settings.SITE_CONFIG_CLIENT, 'Client should be initialized'
    assert settings.SITE_CONFIG_CLIENT.read_only_storage, (
        'GCP storage is initialized')
    assert settings.SITE_CONFIF_CLIENT.cache, 'Cache is initialized'


@mock.patch('google.cloud.storage.Client', mock.Mock())
@pytest.mark.openedx
def test_plugin_production_settings_no_gcp(caplog):
    # Local import to avoid test failures for non-openedx tests
    from site_config_client.openedx.settings.production import plugin_settings

    settings = mock.Mock(
        SITE_CONFIG_CLIENT=None,
        SITE_CONFIG_CACHE_NAME='default',
        SITE_CONFIG_CACHE_TIMEOUT=3600,
        SITE_CONFIG_BASE_URL="http://service",
        SITE_CONFIG_ENVIRONMENT="staging",
        SITE_CONFIG_API_TOKEN="some-token",
        SITE_CONFIG_READ_ONLY_BUCKET=None,
    )

    plugin_settings(settings)

    assert settings.SITE_CONFIG_CLIENT, 'Client should be initialized'
    assert not settings.SITE_CONFIG_CLIENT.read_only_storage, (
        'GCP storage should not be initialized because SITE_CONFIG_READ_ONLY_BUCKET is not provided')
    assert settings.SITE_CONFIF_CLIENT.cache, 'Cache is initialized'

    assert 'Not initializing Google storage' in caplog.text


@pytest.mark.openedx
def test_plugin_production_settings_no_client(caplog):
    # Local import to avoid test failures for non-openedx tests
    from site_config_client.openedx.settings.production import plugin_settings

    settings = mock.Mock(
        SITE_CONFIG_BASE_URL=None,
        SITE_CONFIG_READ_ONLY_BUCKET=None,
    )
    plugin_settings(settings)
    assert settings.SITE_CONFIG_CLIENT, 'Client should not be initialized due to missing SITE_CONFIG_BASE_URL'

    assert 'Not initializing SITE_CONFIG_CLIENT' in caplog.text


@pytest.mark.openedx
def test_plugin_test_settings(client):
    # Local import to avoid test failures for non-openedx tests
    from site_config_client.openedx.settings.test import plugin_settings

    settings = mock.Mock(
        SITE_CONFIG_CLIENT=None,
    )

    plugin_settings(settings)

    assert settings.SITE_CONFIG_CLIENT, 'Client should be initialized'
    assert not settings.SITE_CONFIG_CLIENT.read_only_storage
    assert not settings.SITE_CONFIG_CLIENT.cache



from django.apps import AppConfig

try:
    from openedx.core.djangoapps.plugins.constants import (
        PluginSettings,
        ProjectType,
        SettingsType,
    )
    OPENEDX_ENVIRONMENT = True
except ImportError:  # pragma: no cover
    OPENEDX_ENVIRONMENT = False  # pragma: no cover


class SiteConfigApp(AppConfig):
    """
    Django and Open edX app configs.
    """
    name = 'site_config_client'
    label = 'site_config_client'
    verbose_name = 'Site configuration API client and Open edX plugin.'

    if OPENEDX_ENVIRONMENT:
        # Open edX-specific configurations. This is not used by Django-only environment.
        plugin_app = {
            PluginSettings.CONFIG: {
                ProjectType.LMS: {
                    SettingsType.PRODUCTION: {
                        PluginSettings.RELATIVE_PATH: 'openedx.settings.production'},
                    SettingsType.TEST: {
                        PluginSettings.RELATIVE_PATH: 'openedx.settings.test'},
                },
                ProjectType.CMS: {
                    SettingsType.PRODUCTION: {
                        PluginSettings.RELATIVE_PATH: 'openedx.settings.production'},
                    SettingsType.TEST: {
                        PluginSettings.RELATIVE_PATH: 'openedx.settings.test'},
                },
            },
        }



"""
Tests for GoogleCloudStorage
"""
import logging

from unittest.mock import patch, Mock
from google.api_core.exceptions import NotFound

from site_config_client.google_cloud_storage import GoogleCloudStorage


def get_mock_gcp_client(mock_blob):
    mock_bucket = Mock()
    mock_bucket.get_blob.return_value = mock_blob

    mock_client = Mock()
    mock_client.get_bucket.return_value = mock_bucket
    return mock_client


@patch('google.cloud.storage.Client')
def test_gcp_storage(client_cls):
    mock_blob = Mock()
    mock_blob.download_as_bytes.return_value = b'{"name": "my_site"}'
    client_cls.return_value = get_mock_gcp_client(mock_blob)
    storage = GoogleCloudStorage('random_bucket_name')

    content = storage.read('some_file.json')

    assert content == '{"name": "my_site"}'


@patch('google.cloud.storage.Client')
def test_gcp_storage_not_found(client_cls, caplog):
    caplog.set_level(logging.INFO)
    mock_blob = Mock()
    mock_blob.download_as_bytes.side_effect = NotFound('file not found')
    client_cls.return_value = get_mock_gcp_client(mock_blob)
    storage = GoogleCloudStorage('random_bucket_name')

    content = storage.read('some_file.json')

    assert content is None, \
        'If the file not found, the siteconfg.Client should take of this case'
    assert 'File path not found: some_file.json' in caplog.text


@patch('google.cloud.storage.Client')
def test_gcp_storage_blob_is_none(client_cls, caplog):
    """
    Google Cloud Storage returns a `None` blob, ensure this case is handled.
    """
    caplog.set_level(logging.INFO)
    client_cls.return_value = get_mock_gcp_client(mock_blob=None)
    storage = GoogleCloudStorage('random_bucket_name')

    content = storage.read('some_file.json')

    assert content is None, \
        'If the file not found, the siteconfg.Client should take of this case'
    assert 'File path not found: some_file.json' in caplog.text



"""
Tests for Cache
"""
import pytest


@pytest.fixture
def django_cache_adapter(settings):
    # Local import to avoid test failures for non-openedx tests
    from site_config_client.django_cache import DjangoCache

    settings.CACHES = {
        'default': {
            'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
        }
    }
    return DjangoCache(cache_name='default')


@pytest.mark.django
def test_empty_get(django_cache_adapter):
    cache_key = 'client.040e0ec3-2578-4fcf-b5db-030dadf68f30.live'
    assert django_cache_adapter.get(key=cache_key) is None


@pytest.mark.django
def test_set_get(django_cache_adapter):
    configs = {
     "site": {
            "uuid": "77d4ee4e-6888-4965-b246-b8629ac65bce",
            "domain_name": "croissant.edu",
            "tier": "trial",
            "always_active": False,
            "subscription_ends": "2021-10-31T16:44:45+0000",
            "is_active": True},
     "status": "live",
     "configuration": {
            "admin": [
                {
                    "name": "features",
                    "value": ["bi_connector", "google_tag_manageer"]
                }
            ],
            "css": [
                {
                    "name": "accent-font",
                    "value": "monospace"
                }
            ],
            "page": [],
            "setting": [
                {
                    "name": "PLATFORM_NAME",
                    "value": "Momofuku Academy"
                }
            ],
            "integration": [],
            "secret": [
                {
                    "name": "SEGMENT_KEY",
                    "value": "orc"
                },
                {
                    "name": "SECRET_KEY",
                    "value": "hobbit"
                }
            ]}
    }
    cache_key = 'client.040e0ec3-2578-4fcf-b5db-030dadf68f30.live'
    django_cache_adapter.set(key=cache_key, value=configs)
    assert django_cache_adapter.get(key=cache_key) == configs



"""
Tests for adapter
"""
import pytest


@pytest.mark.django
def test_feature_flag_admin():
    """
    Tests for SiteConfigClientEnabledAdmin.
    """
    from site_config_client.admin import SiteConfigClientEnabledAdmin

    assert SiteConfigClientEnabledAdmin.list_display == ['site_uuid', 'note']
    assert SiteConfigClientEnabledAdmin.search_fields == ['site_uuid', 'note']



"""
Tests for Client
"""
import json

from unittest.mock import Mock
import pytest

from site_config_client.client import Client
from site_config_client.django_cache import DjangoCache
from site_config_client.exceptions import SiteConfigurationError


SITES = {
    "count": 2,
    "results": [
            {
                "uuid": "77d4ee4e-6888-4965-b246-b8629ac65bce",
                "domain_name": "croissant.edu",
                "tier": "trial",
                "always_active": "False",
                "subscription_ends": "2021-10-31T16:44:45+0000",
                "is_active": "True"
            },
            {
                "uuid": "040e0ec3-2578-4fcf-b5db-030dadf68f30",
                "domain_name": "test-site.com",
                "tier": "trial",
                "always_active": "False",
                "subscription_ends": "2021-10-31T16:44:45+0000",
                "is_active": "True"
            }
        ]
}


CONFIGS = {
    "site": {
                "uuid": "77d4ee4e-6888-4965-b246-b8629ac65bce",
                "domain_name": "croissant.edu",
                "tier": "trial",
                "always_active": "False",
                "subscription_ends": "2021-10-31T16:44:45+0000",
                "is_active": "True"
            },
    "status": "draft",
    "configuration":
        {
            "css": [
                    {"name": "accent-font",
                     "value": "monospace"}
                    ],
            "page": [],
            "setting": [
                        {"name": "PLATFORM_NAME",
                         "value": "Momofuku Academy"}
                        ],
            "integration": [],
            "secret": [],
            "admin": [
                      {"name": "features",
                       "value": ["bi_connector", "google_tag_manageer"]}
                    ]
        }
}


SINGLE_CONFIG = {
        "site_uuid": "77d4ee4e-6888-4965-b246-b8629ac65bce",
        "name": "PLATFORM_NAME",
        "value": "Academy of Modern Arts",
        "author_email": "test@example.com",
        "structure_version": "amc-v1",
        "modified": "2021-09-30T16:00:42+0000"
    }


OVERRIDE_CONFIGS = {
        'author_email': 'test@example.com',
        'css': [{
            'name': '$brand-primary-color',
            'value': ['rgba(0,1,1,1)', 'rgba(0,1,1,1)']
        }]
    }


PARAMS = {
    "uuid": "77d4ee4e-6888-4965-b246-b8629ac65bce"
}


def test_client():
    base = "http://127.0.0.1:8000"
    token = "abcaseasyas123"
    bucket = "http://storage.googleapis.com/appsembler/site_config"
    environment = 'staging'
    c = Client(base_url=base,
               api_token=token,
               environment=environment,
               read_only_storage=bucket,
               request_timeout=100,)

    assert c.base_url == base
    assert c.api_token == token
    assert c.environment == environment
    assert c.read_only_storage == bucket
    assert c.request_timeout == 100, 'should not be a tuple'


@pytest.fixture
def site_config_client():
    client = Client(
                base_url="http://service",
                api_token="some-token",
                environment="staging",
    )
    return client


def test_client_has_token(site_config_client):
    assert site_config_client.api_token, 'Client should have an API token'


def test_create_site(site_config_client, requests_mock):
    headers = {'Authorization': '{}'.format(site_config_client.api_token)}
    uuid = 'f12293e5-46b1-46f0'
    site_path = 'http://service/v1/environment/staging/site/'
    requests_mock.post(site_path, json={'site_uuid': uuid},
                       headers=headers, status_code=201)
    new_site = site_config_client.create_site(domain_name='example.com',
                                              site_uuid=uuid)
    assert new_site == {'site_uuid': uuid}
    history = requests_mock.request_history[0]
    assert history.headers.get('Authorization') == 'Token some-token', (
        'API Token passed in Authorization header')


def test_create_site_with_params(site_config_client, requests_mock):
    """
    Tests that uuid is optional, and `params` can be used.
    """
    requests_mock.post(
        'http://service/v1/environment/staging/site/',
        json={},
        headers={'Authorization': '{}'.format(site_config_client.api_token)},
        status_code=201,
    )
    site_config_client.create_site(
        domain_name='example.com',
        params={
            'tier': 'premium',
        }
    )
    history = requests_mock.request_history[0]
    request_json = json.loads(history.body.decode('utf-8'))
    assert request_json == {
        'domain_name': 'example.com',
        'tier': 'premium',
    }


def test_create_site_with_error(site_config_client, requests_mock):
    headers = {'Authorization': '{}'.format(site_config_client.api_token)}

    site_path = 'http://service/v1/environment/staging/site/'
    requests_mock.post(site_path, json={'site_uuid': 'f12293e5-46b1-46f0'},
                       headers=headers, status_code=400)

    with pytest.raises(SiteConfigurationError):
        site_config_client.create_site(domain_name='example.com')

    history = requests_mock.request_history[0]
    assert history.headers.get('Authorization') == 'Token some-token', (
        'API Token passed in Authorization header')


def test_list_sites(site_config_client, requests_mock):
    headers = {'Authorization': '{}'.
               format(site_config_client.api_token)}

    site_path = "http://service/v1/environment/staging/site/"
    requests_mock.get(site_path, json=SITES, headers=headers, status_code=200)
    response_sites = site_config_client.list_sites()
    assert response_sites == SITES
    history = requests_mock.request_history[0]
    assert history.headers.get("Authorization") == "Token some-token", (
        "API Token passed in Authorization header")


def test_list_sites_with_error(site_config_client, requests_mock):
    site_path = "http://service/v1/environment/staging/site/"
    requests_mock.get(site_path, json=SITES, status_code=404)

    with pytest.raises(SiteConfigurationError):
        site_config_client.list_sites()


def test_list_active_sites(site_config_client, requests_mock):
    headers = {'Authorization': '{}'.
               format(site_config_client.api_token)}

    active_sites_path = "http://service/v1/environment/staging/site/?is_active=True"
    requests_mock.get(active_sites_path, json=SITES, headers=headers,
                      status_code=200)
    response_active_sites = site_config_client.list_active_sites()
    assert response_active_sites == SITES
    history = requests_mock.request_history[0]
    assert history.headers.get("Authorization") == "Token some-token", (
        "API Token passed in Authorization header")


def test_list_active_sites_error(site_config_client, requests_mock):
    active_sites_path = "http://service/v1/environment/staging/site/?is_active=True"
    requests_mock.get(active_sites_path, json=SITES, status_code=404)

    with pytest.raises(SiteConfigurationError):
        site_config_client.list_active_sites()


def test_get_backend_configs(requests_mock, site_config_client):
    headers = {'Authorization': '{}'.
               format(site_config_client.api_token)}

    backend_draft_configs_path = (
        'http://service/v1/environment/staging/combined-configuration/backend/{}/draft/'
        .format(PARAMS['uuid']))
    requests_mock.get(backend_draft_configs_path,
                      json=CONFIGS, headers=headers, status_code=200)
    response_configs = site_config_client.get_backend_configs(
        site_uuid=PARAMS['uuid'], status='draft')
    assert response_configs == CONFIGS, (
        'Neither cache nor google cloud storage configured')
    history = requests_mock.request_history[0]
    assert history.headers.get("Authorization") == "Token some-token", (
        "API Token passed in Authorization header")


def test_get_backend_configs_error(requests_mock, site_config_client):
    backend_draft_configs_path = (
        'http://service/v1/environment/staging/combined-configuration/backend/{}/draft/'
        .format(PARAMS['uuid']))
    requests_mock.get(backend_draft_configs_path,
                      json=CONFIGS, status_code=404)

    with pytest.raises(SiteConfigurationError):
        site_config_client.get_backend_configs(
            site_uuid=PARAMS['uuid'], status='draft')


@pytest.mark.django
def test_get_backend_configs_cache(requests_mock, site_config_client):
    site_config_client.cache = DjangoCache(cache_name='default')
    backend_live_configs_path = (
        'http://service/v1/environment/staging/combined-configuration/backend/{}/live/'
        .format(PARAMS['uuid']))
    requests_mock.get(backend_live_configs_path,
                      json=CONFIGS, status_code=200)

    cache_key = 'site_config_client.backend.{}'.format(PARAMS['uuid'])
    config = site_config_client.cache.get(key=cache_key)
    assert config is None, 'Cache does not exist'

    # First read without cache
    fresh_configs = site_config_client.get_backend_configs(
        site_uuid=PARAMS['uuid'], status='live')
    cache_config_value = site_config_client.cache.get(key=cache_key)
    assert cache_config_value == fresh_configs, 'Cache key has been set'

    # Second read with cache
    cached_configs = site_config_client.get_backend_configs(
        site_uuid=PARAMS['uuid'], status='live')
    assert cached_configs == fresh_configs, 'Reading from cache returns the same result'

    # Remove the cache
    site_config_client.delete_cache_for_site(site_uuid=PARAMS['uuid'], status='live')
    assert not site_config_client.get_backend_configs_from_cache(
        site_uuid=PARAMS['uuid'],
        status='live',
    ), 'Should be deleted'


@pytest.mark.django
def test_get_backend_configs_readonly_storage(requests_mock, site_config_client):
    read_only_storage_configs = {
        "site": {
            "uuid": "77d4ee4e-6888-4965-b246-b8629ac65bce",
            "domain_name": "croissant.edu",
            "tier": "trial",
            "always_active": "False",
            "subscription_ends": "2021-10-31T16:44:45+0000",
            "is_active": "True"
        },
        "status": "live",
        "configuration": {},
    }
    storage = Mock()
    storage.read.return_value = json.dumps(read_only_storage_configs)
    site_config_client.read_only_storage = storage
    backend_draft_configs_path = (
        'http://service/v1/environment/staging/combined-configuration/backend/{}/draft/'
        .format(PARAMS['uuid']))
    requests_mock.get(backend_draft_configs_path,
                      json=CONFIGS, status_code=200)

    draft_configs = site_config_client.get_backend_configs(site_uuid=PARAMS['uuid'], status='draft')
    assert draft_configs == CONFIGS, 'Should read draft configs from API directly without using the readonly storage'

    live_configs = site_config_client.get_backend_configs(site_uuid=PARAMS['uuid'], status='live')
    storage.read.assert_called_once_with('v2/{environment}/backend_configs_live_{site_uuid}.json'.format(
        site_uuid=PARAMS['uuid'],
        environment='staging',
    ))
    assert live_configs == read_only_storage_configs


def test_get_config(requests_mock, site_config_client):
    headers = {'Authorization': '{}'.
               format(site_config_client.api_token)}

    config_path = "http://service/v1/environment/staging/configuration/{}/".format(PARAMS['uuid'])
    requests_mock.get(config_path, json=SINGLE_CONFIG, headers=headers,
                      status_code=200)
    response_configs = site_config_client.get_config(
        site_uuid=PARAMS['uuid'],
        type="setting",
        name="PLATFORM_NAME",
        status="live")
    assert response_configs == SINGLE_CONFIG
    history = requests_mock.request_history[0]
    assert history.headers.get("Authorization") == "Token some-token", (
        "API Token passed in Authorization header")


def test_get_config_error(requests_mock, site_config_client):
    config_path = "http://service/v1/environment/staging/configuration/{}/".format(PARAMS['uuid'])

    requests_mock.get(config_path,
                      json=SINGLE_CONFIG, status_code=404)

    with pytest.raises(SiteConfigurationError):
        site_config_client.get_config(
            site_uuid=PARAMS['uuid'],
            type="setting",
            name="PLATFORM_NAME",
            status="live")


def test_override_configs(requests_mock, site_config_client):
    headers = {'Authorization': '{}'.
               format(site_config_client.api_token)}

    override_path = (
        "http://service/v0/environment/staging/configuration-override/{}/"
        .format(PARAMS['uuid']))
    success_response = {
        "developer_message": "Configurations has been overriden successfully."
        }
    requests_mock.put(override_path, json=success_response,
                      headers=headers, status_code=200)
    override_response = site_config_client.override_configs(
        site_uuid=PARAMS['uuid'], configs=OVERRIDE_CONFIGS)
    assert override_response == success_response
    history = requests_mock.request_history[0]
    assert history.headers.get("Authorization") == "Token some-token", (
        "API Token passed in Authorization header")


def test_override_configs_error(requests_mock, site_config_client):
    override_path = (
        "http://service/v0/environment/staging/configuration-override/{}/"
        .format(PARAMS['uuid']))

    requests_mock.put(override_path, status_code=500)

    with pytest.raises(SiteConfigurationError):
        site_config_client.override_configs(
            site_uuid=PARAMS['uuid'], configs=OVERRIDE_CONFIGS)



import json

import requests
from typing import Union
import uuid
from urllib.parse import urljoin

from .exceptions import SiteConfigurationError


STATUS_LIVE = 'live'


def get_cache_key_for_site(site_uuid: Union[str, uuid.UUID]):
    """
    Get the cache key for a site.
    """
    return 'site_config_client.backend.{site_uuid}'.format(
        site_uuid=site_uuid,
    )


class Client:
    def __init__(self, base_url, api_token, environment,
                 read_only_storage=None, cache=None, request_timeout=30):
        """
        Instantiate a new API Client
        """
        self.base_url = base_url
        self.api_token = api_token
        self.environment = environment
        self.read_only_storage = read_only_storage
        self.cache = cache
        self.request_timeout = request_timeout

    def request(self, method, url_path, success_status_code=200, **kwargs):
        """
        Send requests to the Site Configuration service and handle errors.

        Sets timeout and accepts a relative URL.
        """
        headers = {'Authorization': 'Token {}'.format(self.api_token)}
        response = requests.request(
            method=method,
            url=urljoin(self.base_url, url_path),
            timeout=self.request_timeout,
            headers=headers,
            **kwargs
        )

        if response.status_code == success_status_code:
            return response.json()
        else:
            raise SiteConfigurationError((
                'Something went wrong with the site configuration API '
                '`{path}` with status_code="{status_code}" body="{body}"'
            ).format(
                path=url_path,
                status_code=response.status_code,
                body=response.content,
            ))

    def create_site(self, domain_name: str, site_uuid=None, params=None):
        """
        Create a new site.
        """
        params = params or {}
        params['domain_name'] = domain_name
        if site_uuid:
            params['uuid'] = site_uuid
        url = 'v1/environment/{}/site/'.format(self.environment)
        return self.request('post', url, success_status_code=201, json=params)

    def list_sites(self):
        """
        Returns a list of all Sites
        """
        url = 'v1/environment/{}/site/'.format(self.environment)
        return self.request('get', url)

    def list_active_sites(self):
        """
        Returns a list of all active Sites
        """
        url = 'v1/environment/{}/site/?is_active=True'.format(self.environment)
        return self.request('get', url)

    def get_backend_configs_from_readonly_storage(self, site_uuid: Union[str, uuid.UUID], status: str):
        """
        Reads configuration from read-only storage if available for `live` status only.
        """
        if status != STATUS_LIVE:
            # The read-only storage makes sense only for live (published) configs.
            return None

        file_path = 'v2/{environment}/backend_configs_live_{site_uuid}.json'.format(
            environment=self.environment,
            site_uuid=site_uuid,
        )

        config_str = None
        if self.read_only_storage:
            config_str = self.read_only_storage.read(file_path)

        if config_str:
            return json.loads(config_str)

        return None

    def get_backend_configs_from_api(self, site_uuid: Union[str, uuid.UUID], status: str):
        """
        Get the backend config from the Site Configuration Service API.
        """
        api_endpoint = 'v1/environment/{}/combined-configuration/backend/{}/{}/'.format(
            self.environment, site_uuid, status
        )
        return self.request('get', url_path=api_endpoint)

    def get_backend_configs_from_cache(self, site_uuid: Union[str, uuid.UUID], status: str):
        """
        Get the backend config from the cache -- if available.
        """
        config = None
        if self.cache and status == STATUS_LIVE:
            # Only cache live status configs. Draft should always be fetched fresh.
            config = self.cache.get(key=get_cache_key_for_site(site_uuid))
        return config

    def set_backend_configs_in_cache(self, site_uuid: Union[str, uuid.UUID], status: str, config):
        """
        Store the config in cache.
        """
        if self.cache and status == STATUS_LIVE:
            # Only cache live status configs. Draft should always be fetched fresh.
            self.cache.set(key=get_cache_key_for_site(site_uuid), value=config)

    def delete_cache_for_site(self, site_uuid: Union[str, uuid.UUID], status):
        """
        Clear cache entry for a specific site.
        """
        if self.cache and status == STATUS_LIVE:
            # Only cache live status configs. Draft should always be fetched fresh.
            self.cache.delete(key=get_cache_key_for_site(site_uuid))

    def get_backend_configs(self, site_uuid: Union[str, uuid.UUID], status: str):
        """
        Returns a combination of Site information and `live` or `draft`
        Configurations (backend secrets included)

        [Client Configuration]
        - Django Cache
            - if cache key exists: return config from cache
            - if cache key does not exist: call endpoint to get config, set
              cache with config, return config
        """
        config = self.get_backend_configs_from_cache(site_uuid, status)

        if config:
            store_in_cache = False
        else:
            store_in_cache = True
            config = self.get_backend_configs_from_readonly_storage(site_uuid, status)

        if not config:
            config = self.get_backend_configs_from_api(site_uuid, status)

        if store_in_cache:
            self.set_backend_configs_in_cache(site_uuid, status, config)

        return config

    def get_config(self, site_uuid: Union[str, uuid.UUID],
                   type: str, name: str, status: str):
        """
        Returns a single configuration object for Site
        """
        api_endpoint = 'v1/environment/{}/configuration/{}/'.format(self.environment, site_uuid)
        return self.request('get', url_path=api_endpoint, json={
            "type": type,
            "name": name,
            "status": status
        })

    def override_configs(self, site_uuid: Union[str, uuid.UUID], configs):
        """
        Override all live configs in a single pass.

        This uses the v0 API which should be deprecated after the initial
        rollout.
        """
        api_endpoint = 'v0/environment/{environment}/configuration-override/{site_uuid}/'.format(
            environment=self.environment,
            site_uuid=site_uuid,
        )
        return self.request('put', url_path=api_endpoint, json=configs)



"""
Test settings for the Open edX
"""

from site_config_client.client import Client


def plugin_settings(settings):
    """
    Initialize a client that won't work. Useful for testing, but still needs mocking when used.
    """
    api_token = 'not real token'  # nosec

    settings.SITE_CONFIG_CLIENT = Client(
        base_url='http://localhost:14000/v1/client/',
        api_token=api_token,
        environment='test',
    )



"""
Tests for adapter
"""

from uuid import uuid4
import pytest


@pytest.mark.openedx
def test_global_feature_flag(settings):
    from site_config_client.openedx.features import is_feature_enabled
    assert not is_feature_enabled(), 'Should be disabled by default'

    settings.FEATURES = {'SITE_CONFIG_CLIENT_ENABLED': True}
    assert is_feature_enabled(), 'Should be enabled because of feature flag'


@pytest.mark.openedx
@pytest.mark.django_db
def test_site_feature_flags_with_global_flag(settings):
    from site_config_client.openedx.features import is_feature_enabled_for_site
    site_uuid = uuid4()

    assert not is_feature_enabled_for_site(site_uuid), 'Should be disabled by default'

    settings.FEATURES = {'SITE_CONFIG_CLIENT_ENABLED': True}
    assert is_feature_enabled_for_site(site_uuid), 'Should be enabled because of global feature flag'


@pytest.mark.openedx
@pytest.mark.django_db
def test_site_feature_flags_without_global_flag():
    from site_config_client.models import SiteConfigClientEnabled
    from site_config_client.openedx import features
    site_uuid = uuid4()

    assert not features.is_feature_enabled_for_site(site_uuid), 'Should be disabled by default'
    assert not SiteConfigClientEnabled.objects.filter(site_uuid=site_uuid).exists()

    features.enable_feature_for_site(site_uuid=site_uuid)

    assert SiteConfigClientEnabled.objects.filter(site_uuid=site_uuid).exists(), 'enable_feature_for_site adds it'
    assert features.is_feature_enabled_for_site(site_uuid), 'Should be enabled because of site-specific feature flag'


@pytest.mark.openedx
@pytest.mark.django_db
def test_enable_feature_for_site_with_note():
    from site_config_client.models import SiteConfigClientEnabled
    from site_config_client.openedx import features

    site_uuid = uuid4()
    assert features.enable_feature_for_site(site_uuid=site_uuid, note='my note'), 'should create new entry'
    assert not features.enable_feature_for_site(site_uuid=site_uuid), 'should keep the old entry'
    assert features.is_feature_enabled_for_site(site_uuid), 'calling the helper twice do not hurt'
    flag = SiteConfigClientEnabled.objects.get(site_uuid=site_uuid)
    assert flag.note == 'my note', 'enable_feature_for_site can add a note'



"""
Production settings for the Open edX
"""
import logging

from site_config_client.client import Client
from site_config_client.django_cache import DjangoCache


log = logging.getLogger(__name__)


def plugin_settings(settings):
    """
    Production and devstack settings for Site Configuration API Client.
    """

    bucket_name = getattr(settings, 'SITE_CONFIG_READ_ONLY_BUCKET', None)
    read_only_storage = None

    if bucket_name:
        # Google Cloud Storage is optional and used for service reliability
        from site_config_client.google_cloud_storage import GoogleCloudStorage
        read_only_storage = GoogleCloudStorage(
            bucket_name=settings.SITE_CONFIG_READ_ONLY_BUCKET,
        )
    else:
        log.warning('Not initializing Google storage bucket due to missing settings parameter.')

    cache = DjangoCache(
        cache_name=getattr(settings, 'SITE_CONFIG_CACHE_NAME', 'default'),
        cache_timeout=getattr(settings, 'SITE_CONFIG_CACHE_TIMEOUT', None),
    )

    base_url = getattr(settings, 'SITE_CONFIG_BASE_URL', None)
    if base_url:
        settings.SITE_CONFIG_CLIENT = Client(
            base_url=settings.SITE_CONFIG_BASE_URL,
            api_token=settings.SITE_CONFIG_API_TOKEN,
            environment=settings.SITE_CONFIG_ENVIRONMENT,
            read_only_storage=read_only_storage,
            cache=cache,
        )
    else:
        log.warning('Not initializing SITE_CONFIG_CLIENT due to missing parameter.')



"""
Feature helpers for Open edX.
"""

from django.conf import settings

from site_config_client.models import SiteConfigClientEnabled


def is_feature_enabled():
    """
    Checks if the feature is globally enabled.
    """
    enabled_globally = settings.FEATURES.get('SITE_CONFIG_CLIENT_ENABLED', False)
    return enabled_globally


def enable_feature_for_site(site_uuid, note=''):
    """
    Enable the site configuration client for a site by site_uuid.

    Returns: whether a new flag is created or not.
    """
    _, created = SiteConfigClientEnabled.objects.get_or_create(
        site_uuid=site_uuid,
        defaults={
            'note': note,
        }
    )
    return created


def is_feature_enabled_for_site(site_uuid):
    """
    Checks if feature is globally enabled OR for the specific site.

    This helps to facilitates gradual migration.
    """
    if is_feature_enabled():
        return True

    return SiteConfigClientEnabled.objects.filter(site_uuid=site_uuid).exists()



"""
Tests for adapter
"""
import pytest
from unittest.mock import Mock


CONFIGS = {
    "site": {
        "uuid": "77d4ee4e-6888-4965-b246-b8629ac65bce",
        "always_active": True,
        "subscription_ends": "2021-11-12T15:30:43+0000",
        "tier": "trial",
    },
    "status": "live",
    "configuration": {
        "css": {
            "selected_font": "lato",
            "text_color": "#0a0a0a",
            "header_logo_height": "110",
        },
        "page": {
            "course-card": "course-tile-01",
            "privacy": {
                "content": [
                    {
                     "element-type": "layout-50:50",
                     "element-path": "page-builder/layouts/_two-col-50-50.html"
                    }
                ]
            }
        },
        "setting": {
            "PLATFORM_NAME": "My New Platform Name!",
            "footer_copyright_text": "© Appsembler 2021. All rights reserved.",
            "display_footer_powered_by": "false",
            "display_footer_legal": "false",
            "google_verification_code": "GoogleVerify!",
            "site_title": "My site title"
        },
        "integration": {},
        "secret": {
            "SEGMENT_KEY": "so secret",
        },
        "admin": {},
        }
}


@pytest.mark.openedx
def test_adapater(settings):
    '''
    I want to mock the return value for `get_backend_config` as CONFIGS
    and test the return values of:
     `get_value`,  `get_amc_v1_theme_css_variables`, `get_amc_v1_page`
    '''
    from site_config_client.openedx.adapter import SiteConfigAdapter

    mock = Mock()
    mock.get_backend_configs.return_value = CONFIGS
    settings.SITE_CONFIG_CLIENT = mock
    adapter = SiteConfigAdapter("77d4ee4e-6888-4965", "live")

    assert adapter.get_backend_configs() == CONFIGS

    setting_platform_name = adapter.get_value_of_type('setting', 'PLATFORM_NAME', None)
    assert setting_platform_name == CONFIGS['configuration']['setting']['PLATFORM_NAME']
    assert adapter.get_value_of_type('secret', 'SEGMENT_KEY', None) == 'so secret'

    css_vars = adapter.get_css_variables_dict()
    assert css_vars == {
        "selected_font": "lato",
        "text_color": "#0a0a0a",
        "header_logo_height": "110",
    }, 'get_css_variables_dict returns a dictionary of all variables'

    privacy_page_vars = adapter.get_value_of_type('page', 'privacy', None)
    assert privacy_page_vars == CONFIGS['configuration']['page']['privacy']

    assert adapter.get_value_of_type('page', 'non_existent', None) is None, 'Should default to None'
    assert adapter.get_value_of_type('page', 'non_existent', {'title': 'default'}) == {
        'title': 'default',
    }, 'Should has a default'

    assert adapter.backend_configs, 'Sanity check: in-memory cache should be available'
    adapter.delete_backend_configs_cache()
    assert not adapter.backend_configs, 'Should remove in-memory cache'

    site_info = adapter.get_site_info()
    assert site_info['subscription_ends'].year == 2021, 'should return a datetime object'



import os
from setuptools import find_packages, setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='site-configuration-client',
    version='0.2.2',
    description='Python client library for Site Configuration API',
    long_description=read('README.rst'),
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(),
    install_requires=[
        "requests>=2.20.0",
    ],
    entry_points={
        'lms.djangoapp': [
            'site_config_client = site_config_client.apps:SiteConfigApp',
        ],
        'cms.djangoapp': [
            'site_config_client = site_config_client.apps:SiteConfigApp',
        ],
    },
    url="https://github.com/appsembler/site-configuration-client"
)



"""
Models for the Site Configuration client.
"""

from django.contrib import admin

from .models import SiteConfigClientEnabled


@admin.register(SiteConfigClientEnabled)
class SiteConfigClientEnabledAdmin(admin.ModelAdmin):
    """
    Admin interface for SiteConfigClientEnabled.
    """

    list_display = ['site_uuid', 'note']
    search_fields = ['site_uuid', 'note']
